@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -37,13 +37,18 @@
  */
 
 #include "rsh_locl.h"
-RCSID("$Id: rsh.c,v 1.15 1997/12/17 01:31:27 assar Exp $");
+RCSID("$Id: rsh.c,v 1.37 1999/08/04 17:03:54 assar Exp $");
 
 enum auth_method auth_method;
 int do_encrypt;
 int do_forward;
+int do_forwardable;
+int do_unique_tkfile = 0;
+char *unique_tkfile  = NULL;
+char tkfile[MAXPATHLEN];
 krb5_context context;
 krb5_keyblock *keyblock;
+krb5_crypto crypto;
 des_key_schedule schedule;
 des_cblock iv;
 
@@ -52,20 +57,22 @@ des_cblock iv;
  *
  */
 
-static int input;		/* Read from stdin */
+static int input = 1;		/* Read from stdin */
 
 static int
 loop (int s, int errsock)
 {
     fd_set real_readset;
-    int count = 2;
+    int count = 1;
 
     FD_ZERO(&real_readset);
     FD_SET(s, &real_readset);
-    FD_SET(errsock, &real_readset);
-    if(input) {
-	FD_SET(STDIN_FILENO, &real_readset);
+    if (errsock != -1) {
+	FD_SET(errsock, &real_readset);
+	++count;
     }
+    if(input)
+	FD_SET(STDIN_FILENO, &real_readset);
 
     for (;;) {
 	int ret;
@@ -74,11 +81,12 @@ loop (int s, int errsock)
 
 	readset = real_readset;
 	ret = select (max(s, errsock) + 1, &readset, NULL, NULL, NULL);
-	if (ret < 0)
+	if (ret < 0) {
 	    if (errno == EINTR)
 		continue;
 	    else
 		err (1, "select");
+	}
 	if (FD_ISSET(s, &readset)) {
 	    ret = do_read (s, buf, sizeof(buf));
 	    if (ret < 0)
@@ -91,7 +99,7 @@ loop (int s, int errsock)
 	    } else
 		net_write (STDOUT_FILENO, buf, ret);
 	}
-	if (FD_ISSET(errsock, &readset)) {
+	if (errsock != -1 && FD_ISSET(errsock, &readset)) {
 	    ret = do_read (errsock, buf, sizeof(buf));
 	    if (ret < 0)
 		err (1, "read");
@@ -110,6 +118,7 @@ loop (int s, int errsock)
 	    else if (ret == 0) {
 		close (STDIN_FILENO);
 		FD_CLR(STDIN_FILENO, &real_readset);
+		shutdown (s, SHUT_WR);
 	    } else
 		do_write (s, buf, ret);
 	}
@@ -117,15 +126,15 @@ loop (int s, int errsock)
 }
 
 #ifdef KRB4
-static void
+static int
 send_krb4_auth(int s,
-	       struct sockaddr_in thisaddr,
-	       struct sockaddr_in thataddr,
-	       char *hostname,
-	       char *remote_user,
-	       char *local_user,
+	       struct sockaddr *thisaddr,
+	       struct sockaddr *thataddr,
+	       const char *hostname,
+	       const char *remote_user,
+	       const char *local_user,
 	       size_t cmd_len,
-	       char *cmd)
+	       const char *cmd)
 {
     KTEXT_ST text;
     CREDENTIALS cred;
@@ -135,25 +144,40 @@ send_krb4_auth(int s,
 
     status = krb_sendauth (do_encrypt ? KOPT_DO_MUTUAL : 0,
 			   s, &text, "rcmd",
-			   hostname, krb_realmofhost (hostname),
+			   (char *)hostname, krb_realmofhost (hostname),
 			   getpid(), &msg, &cred, schedule,
-			   &thisaddr, &thataddr, KCMD_VERSION);
-    if (status != KSUCCESS)
-	errx (1, "%s: %s", hostname, krb_get_err_text(status));
+			   (struct sockaddr_in *)thisaddr,
+			   (struct sockaddr_in *)thataddr,
+			   KCMD_VERSION);
+    if (status != KSUCCESS) {
+	warnx ("%s: %s", hostname, krb_get_err_text(status));
+	return 1;
+    }
     memcpy (iv, cred.session, sizeof(iv));
 
     len = strlen(remote_user) + 1;
-    if (net_write (s, remote_user, len) != len)
-	err (1, "write");
-    if (net_write (s, cmd, cmd_len) != cmd_len)
-	err (1, "write");
+    if (net_write (s, remote_user, len) != len) {
+	warn("write");
+	return 1;
+    }
+    if (net_write (s, cmd, cmd_len) != cmd_len) {
+	warn("write");
+	return 1;
+    }
+    return 0;
 }
 #endif /* KRB4 */
 
+/*
+ * Send forward information on `s' for host `hostname', them being
+ * forwardable themselves if `forwardable'
+ */
+
 static int
 krb5_forward_cred (krb5_auth_context auth_context,
 		   int s,
-		   char *hostname)
+		   const char *hostname,
+		   int forwardable)
 {
     krb5_error_code ret;
     krb5_ccache     ccache;
@@ -162,6 +186,8 @@ krb5_forward_cred (krb5_auth_context auth_context,
     krb5_data       out_data;
     krb5_principal  principal;
 
+    memset (&creds, 0, sizeof(creds));
+
     ret = krb5_cc_default (context, &ccache);
     if (ret) {
 	warnx ("could not forward creds: krb5_cc_default: %s",
@@ -195,7 +221,8 @@ krb5_forward_cred (krb5_auth_context auth_context,
     creds.times.endtime = 0;
 
     flags.i = 0;
-    flags.b.forwarded = 1;
+    flags.b.forwarded   = 1;
+    flags.b.forwardable = forwardable;
 
     ret = krb5_get_forwarded_creds (context,
 				    auth_context,
@@ -221,38 +248,35 @@ krb5_forward_cred (krb5_auth_context auth_context,
     return 0;
 }
 
-static void
+static int
 send_krb5_auth(int s,
-	       struct sockaddr_in thisaddr,
-	       struct sockaddr_in thataddr,
-	       char *hostname,
-	       char *remote_user,
-	       char *local_user,
+	       struct sockaddr *thisaddr,
+	       struct sockaddr *thataddr,
+	       const char *hostname,
+	       const char *remote_user,
+	       const char *local_user,
 	       size_t cmd_len,
-	       char *cmd)
+	       const char *cmd)
 {
     krb5_principal server;
     krb5_data cksum_data;
-    krb5_ccache ccache;
     int status;
     size_t len;
     krb5_auth_context auth_context = NULL;
 
-    krb5_init_context(&context);
-
-    krb5_cc_default (context, &ccache);
-
     status = krb5_sname_to_principal(context,
 				     hostname,
 				     "host",
 				     KRB5_NT_SRV_HST,
 				     &server);
-    if (status)
-	errx (1, "%s: %s", hostname, krb5_get_err_text(context, status));
+    if (status) {
+	warnx ("%s: %s", hostname, krb5_get_err_text(context, status));
+	return 1;
+    }
 
     cksum_data.length = asprintf ((char **)&cksum_data.data,
 				  "%u:%s%s%s",
-				  ntohs(thataddr.sin_port),
+				  ntohs(socket_get_port(thataddr)),
 				  do_encrypt ? "-x " : "",
 				  cmd,
 				  remote_user);
@@ -266,138 +290,207 @@ send_krb5_auth(int s,
 			    do_encrypt ? AP_OPTS_MUTUAL_REQUIRED : 0,
 			    &cksum_data,
 			    NULL,
-			    ccache,
+			    NULL,
 			    NULL,
 			    NULL,
 			    NULL);
-    if (status)
-	errx (1, "%s: %s", hostname, krb5_get_err_text(context, status));
+    if (status) {
+	warnx ("%s: %s", hostname, krb5_get_err_text(context, status));
+	return 1;
+    }
 
     status = krb5_auth_con_getkey (context, auth_context, &keyblock);
-    if (status)
-      errx (1, "krb5_auth_con_getkey: %s",
-	    krb5_get_err_text(context, status));
+    if (status) {
+	warnx ("krb5_auth_con_getkey: %s", krb5_get_err_text(context, status));
+	return 1;
+    }
+
+    status = krb5_auth_con_setaddrs_from_fd (context,
+					     auth_context,
+					     &s);
+    if (status) {
+        warnx("krb5_auth_con_setaddrs_from_fd: %s",
+	      krb5_get_err_text(context, status));
+        return(1);
+    }
+
+    status = krb5_crypto_init(context, keyblock, 0, &crypto);
+    if(status) {
+	warnx ("krb5_crypto_init: %s", krb5_get_err_text(context, status));
+	return 1;
+    }
 
     len = strlen(remote_user) + 1;
-    if (net_write (s, remote_user, len) != len)
-	err (1, "write");
-    if (do_encrypt && net_write (s, "-x ", 3) != 3)
-	err (1, "write");
-    if (net_write (s, cmd, cmd_len) != cmd_len)
-	err (1, "write");
-    len = strlen(local_user) + 1;
-    if (net_write (s, local_user, len) != len)
-	err (1, "write");
+    if (net_write (s, remote_user, len) != len) {
+	warn ("write");
+	return 1;
+    }
+    if (do_encrypt && net_write (s, "-x ", 3) != 3) {
+	warn ("write");
+	return 1;
+    }
+    if (net_write (s, cmd, cmd_len) != cmd_len) {
+	warn ("write");
+	return 1;
+    }
 
-    if (!do_forward || krb5_forward_cred (auth_context, s, hostname)) {
+    if (do_unique_tkfile) {
+	if (net_write (s, tkfile, strlen(tkfile)) != strlen(tkfile)) {
+	    warn ("write");
+	    return 1;
+	}
+    }
+    len = strlen(local_user) + 1;
+    if (net_write (s, local_user, len) != len) {
+	warn ("write");
+	return 1;
+    }
+
+    if (!do_forward
+	|| krb5_forward_cred (auth_context, s, hostname, do_forwardable)) {
 	/* Empty forwarding info */
 
 	u_char zero[4] = {0, 0, 0, 0};
 	write (s, &zero, 4);
     }
     krb5_auth_con_free (context, auth_context);
-
+    return 0;
 }
 
-static void
+static int
 send_broken_auth(int s,
-		 struct sockaddr_in thisaddr,
-		 struct sockaddr_in thataddr,
-		 char *hostname,
-		 char *remote_user,
-		 char *local_user,
+		 struct sockaddr *thisaddr,
+		 struct sockaddr *thataddr,
+		 const char *hostname,
+		 const char *remote_user,
+		 const char *local_user,
 		 size_t cmd_len,
-		 char *cmd)
+		 const char *cmd)
 {
     size_t len;
 
     len = strlen(local_user) + 1;
-    if (net_write (s, local_user, len) != len)
-	err (1, "write");
+    if (net_write (s, local_user, len) != len) {
+	warn ("write");
+	return 1;
+    }
     len = strlen(remote_user) + 1;
-    if (net_write (s, remote_user, len) != len)
-	err (1, "write");
-    if (net_write (s, cmd, cmd_len) != cmd_len)
-	err (1, "write");
+    if (net_write (s, remote_user, len) != len) {
+	warn ("write");
+	return 1;
+    }
+    if (net_write (s, cmd, cmd_len) != cmd_len) {
+	warn ("write");
+	return 1;
+    }
+    return 0;
 }
 
 static int
 proto (int s, int errsock,
-       char *hostname, char *local_user, char *remote_user,
-       char *cmd, size_t cmd_len)
+       const char *hostname, const char *local_user, const char *remote_user,
+       const char *cmd, size_t cmd_len,
+       int (*auth_func)(int s,
+			struct sockaddr *this, struct sockaddr *that,
+			const char *hostname, const char *remote_user,
+			const char *local_user, size_t cmd_len,
+			const char *cmd))
 {
-    struct sockaddr_in erraddr;
     int errsock2;
     char buf[BUFSIZ];
     char *p;
     size_t len;
     char reply;
-    struct sockaddr_in thisaddr, thataddr;
+    struct sockaddr_storage thisaddr_ss;
+    struct sockaddr *thisaddr = (struct sockaddr *)&thisaddr_ss;
+    struct sockaddr_storage thataddr_ss;
+    struct sockaddr *thataddr = (struct sockaddr *)&thataddr_ss;
+    struct sockaddr_storage erraddr_ss;
+    struct sockaddr *erraddr = (struct sockaddr *)&erraddr_ss;
     int addrlen;
     int ret;
 
-    addrlen = sizeof(thisaddr);
-    if (getsockname (s, (struct sockaddr *)&thisaddr, &addrlen) < 0
-	|| addrlen != sizeof(thisaddr)) {
-	err (1, "getsockname(%s)", hostname);
+    addrlen = sizeof(thisaddr_ss);
+    if (getsockname (s, thisaddr, &addrlen) < 0) {
+	warn ("getsockname(%s)", hostname);
+	return 1;
     }
-    addrlen = sizeof(thataddr);
-    if (getpeername (s, (struct sockaddr *)&thataddr, &addrlen) < 0
-	|| addrlen != sizeof(thataddr)) {
-	err (1, "getpeername(%s)", hostname);
+    addrlen = sizeof(thataddr_ss);
+    if (getpeername (s, thataddr, &addrlen) < 0) {
+	warn ("getpeername(%s)", hostname);
+	return 1;
     }
 
-    addrlen = sizeof(erraddr);
-    if (getsockname (errsock, (struct sockaddr *)&erraddr, &addrlen) < 0)
-	err (1, "getsockname");
+    if (errsock != -1) {
 
-    if (listen (errsock, 1) < 0)
-	err (1, "listen");
+	addrlen = sizeof(erraddr_ss);
+	if (getsockname (errsock, erraddr, &addrlen) < 0) {
+	    warn ("getsockname");
+	    return 1;
+	}
 
-    p = buf;
-    snprintf (p, sizeof(buf), "%u", ntohs(erraddr.sin_port));
-    len = strlen(buf) + 1;
-    if(net_write (s, buf, len) != len)
-	err (1, "write");
+	if (listen (errsock, 1) < 0) {
+	    warn ("listen");
+	    return 1;
+	}
 
-    errsock2 = accept (errsock, NULL, NULL);
-    if (errsock2 < 0)
-	err (1, "accept");
-    close (errsock);
+	p = buf;
+	snprintf (p, sizeof(buf), "%u",
+		  ntohs(socket_get_port(erraddr)));
+	len = strlen(buf) + 1;
+	if(net_write (s, buf, len) != len) {
+	    warn ("write");
+	    close (errsock);
+	    return 1;
+	}
 
-#ifdef KRB4
-    if (auth_method == AUTH_KRB4)
-	send_krb4_auth (s, thisaddr, thataddr,
-			hostname, remote_user, local_user,
-			cmd_len, cmd);
-    else
-#endif /* KRB4 */
-    if(auth_method == AUTH_KRB5)
-	send_krb5_auth (s, thisaddr, thataddr,
-			hostname, remote_user, local_user,
-			cmd_len, cmd);
-    else
-    if(auth_method == AUTH_BROKEN)
-	send_broken_auth (s, thisaddr, thataddr,
-			  hostname, remote_user, local_user,
-			  cmd_len, cmd);
-    else
-	abort ();
+	errsock2 = accept (errsock, NULL, NULL);
+	if (errsock2 < 0) {
+	    warn ("accept");
+	    close (errsock);
+	    return 1;
+	}
+	close (errsock);
 
-    if (net_read (s, &reply, 1) != 1)
-	err (1, "read");
+    } else {
+	if (net_write (s, "0", 2) != 2) {
+	    warn ("write");
+	    return 1;
+	}
+	errsock2 = -1;
+    }
+
+    if ((*auth_func)(s, thisaddr, thataddr, hostname,
+		     remote_user, local_user,
+		     cmd_len, cmd)) {
+	close (errsock2);
+	return 1;
+    } 
+
+    if (net_read (s, &reply, 1) != 1) {
+	warn ("read");
+	close (errsock2);
+	return 1;
+    }
     if (reply != 0) {
 
 	warnx ("Error from rshd at %s:", hostname);
 
 	while ((ret = read (s, buf, sizeof(buf))) > 0)
 	    write (STDOUT_FILENO, buf, ret);
+        write (STDOUT_FILENO,"\n",1);
+	close (errsock2);
 	return 1;
     }
 
     return loop (s, errsock2);
 }
 
+/*
+ * Return in `res' a copy of the concatenation of `argc, argv' into
+ * malloced space.
+ */
+
 static size_t
 construct_command (char **res, int argc, char **argv)
 {
@@ -407,6 +500,7 @@ construct_command (char **res, int argc, char **argv)
 
     for (i = 0; i < argc; ++i)
 	len += strlen(argv[i]) + 1;
+    len = max (1, len);
     tmp = malloc (len);
     if (tmp == NULL)
 	errx (1, "malloc %u failed", len);
@@ -416,42 +510,73 @@ construct_command (char **res, int argc, char **argv)
 	strcat (tmp, argv[i]);
 	strcat (tmp, " ");
     }
-    strcat (tmp, argv[argc-1]);
+    if (argc > 0)
+	strcat (tmp, argv[argc-1]);
     *res = tmp;
     return len;
 }
 
+static char *
+print_addr (int af, const void *a)
+{
+    char addr_str[256];
+    char *res;
+
+    inet_ntop (af, a, addr_str, sizeof(addr_str));
+    res = strdup(addr_str);
+    if (res == NULL)
+	errx (1, "malloc: out of memory");
+    return res;
+}
+
 static int
 doit_broken (int argc,
 	     char **argv,
 	     int optind,
-	     char *remote_user,
-	     char *local_user,
+	     const char *host,
+	     const char *remote_user,
+	     const char *local_user,
 	     int port,
 	     int priv_socket1,
 	     int priv_socket2,
-	     char *cmd,
+	     const char *cmd,
 	     size_t cmd_len)
 {
-    struct hostent *hostent;
-    struct sockaddr_in addr;
+    struct hostent *hostent = NULL;
+    struct sockaddr_storage addr_ss;
+    struct sockaddr *addr = (struct sockaddr *)&addr_ss;
+    int error;
+    int af;
 
-    if (priv_socket1 < 0 || priv_socket2 < 0)
-	errx (1, "unable to bind reserved port: is rsh setuid root?");
+    if (priv_socket1 < 0) {
+	warnx ("unable to bind reserved port: is rsh setuid root?");
+	return 1;
+    }
 
-    hostent = roken_gethostbyname (argv[optind]);
+#ifdef HAVE_IPV6    
     if (hostent == NULL)
-	errx (1, "gethostbyname '%s' failed: %s",
-	      argv[optind],
-	      hstrerror(h_errno));
+	hostent = getipnodebyname (host, AF_INET6, 0, &error);
+#endif
+    if (hostent == NULL)
+	hostent = getipnodebyname (host, AF_INET, 0, &error);
 
+    if (hostent == NULL) {
+	warn("gethostbyname '%s' failed: %s", host, hstrerror(error));
+	return 1;
+    }
+
+    af = addr->sa_family = hostent->h_addrtype;
+    socket_set_address_and_port (addr, hostent->h_addr_list[0], port);
+
+#if 0
     memset (&addr, 0, sizeof(addr));
     addr.sin_family = AF_INET;
     addr.sin_port   = port;
     addr.sin_addr   = *((struct in_addr *)hostent->h_addr_list[0]);
+#endif
 
-    if (connect(priv_socket1, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
-	struct in_addr **h;
+    if (connect(priv_socket1, addr, socket_sockaddr_size(addr)) < 0) {
+	char **h;
 
 	if (hostent->h_addr_list[1] == NULL)
 	    return 1;
@@ -459,9 +584,7 @@ doit_broken (int argc,
 	close(priv_socket1);
 	close(priv_socket2);
 
-	for(h = (struct in_addr **)hostent->h_addr_list;
-	    *h != NULL;
-	    ++h) {
+	for(h = hostent->h_addr_list; *h != NULL; ++h) {
 	    pid_t pid;
 
 	    pid = fork();
@@ -469,22 +592,28 @@ doit_broken (int argc,
 		err (1, "fork");
 	    else if(pid == 0) {
 		char **new_argv;
-		int i;
+		int i = 0;
 
 		new_argv = malloc((argc + 2) * sizeof(*new_argv));
 		if (new_argv == NULL)
 		    errx (1, "malloc: out of memory");
-		new_argv[0] = argv[0];
-		new_argv[1] = "-K";
-		for(i = 1; i < argc; ++i)
-		    new_argv[i + 1] = argv[i];
-		new_argv[optind + 1] = inet_ntoa(**h);
+		new_argv[i] = argv[i];
+		++i;
+		if (optind == i)
+		    new_argv[i++] = print_addr (af, *h);
+		new_argv[i++] = "-K";
+		for(; i <= argc; ++i)
+		    new_argv[i] = argv[i - 1];
+		if (optind > 1)
+		    new_argv[optind + 1] = print_addr(af, *h);
 		new_argv[argc + 1] = NULL;
 		execv(PATH_RSH, new_argv);
 		err(1, "execv(%s)", PATH_RSH);
 	    } else {
 		int status;
 
+		freehostent (hostent);
+
 		while(waitpid(pid, &status, 0) < 0)
 		    ;
 		if(WIFEXITED(status) && WEXITSTATUS(status) == 0)
@@ -493,63 +622,94 @@ doit_broken (int argc,
 	}
 	return 1;
     } else {
-	return proto (priv_socket1, priv_socket2,
-		      argv[optind],
-		      local_user, remote_user,
-		      cmd, cmd_len);
+	int ret;
+
+	freehostent (hostent);
+
+	ret = proto (priv_socket1, priv_socket2,
+		     argv[optind],
+		     local_user, remote_user,
+		     cmd, cmd_len,
+		     send_broken_auth);
+	return ret;
     }
 }
 
 static int
-doit (char *hostname,
-      char *remote_user,
-      char *local_user,
+doit (const char *hostname,
+      const char *remote_user,
+      const char *local_user,
       int port,
-      char *cmd,
-      size_t cmd_len)
+      const char *cmd,
+      size_t cmd_len,
+      int do_errsock,
+      int (*auth_func)(int s,
+		       struct sockaddr *this, struct sockaddr *that,
+		       const char *hostname, const char *remote_user,
+		       const char *local_user, size_t cmd_len,
+		       const char *cmd))
 {
-    struct hostent *hostent;
-    struct in_addr **h;
+    struct hostent *hostent = NULL;
+    int error;
+    char **h;
+    int af;
+
+#ifdef HAVE_IPV6    
+    if (hostent == NULL)
+	hostent = getipnodebyname (hostname, AF_INET6, 0, &error);
+#endif
+    if (hostent == NULL)
+	hostent = getipnodebyname (hostname, AF_INET, 0, &error);
 
-    hostent = roken_gethostbyname (hostname);
     if (hostent == NULL)
 	errx (1, "gethostbyname '%s' failed: %s",
 	      hostname,
-	      hstrerror(h_errno));
-    for (h = (struct in_addr **)hostent->h_addr_list;
-	*h != NULL;
-	 ++h) {
+	      hstrerror(error));
+    af = hostent->h_addrtype;
+    for (h = hostent->h_addr_list; *h != NULL; ++h) {
 	int s;
-	struct sockaddr_in addr;
+	struct sockaddr_storage addr_ss;
+	struct sockaddr *addr = (struct sockaddr *)&addr_ss;
 	int errsock;
-	struct sockaddr_in erraddr;
+	struct sockaddr_storage erraddr_ss;
+	struct sockaddr *erraddr = (struct sockaddr *)&erraddr_ss;
+	int ret;
 
+	addr->sa_family = af;
+	socket_set_address_and_port (addr, *h, port);
+
+#if 0
 	memset (&addr, 0, sizeof(addr));
 	addr.sin_family = AF_INET;
 	addr.sin_port   = port;
 	addr.sin_addr   = **h;
+#endif
 
-	s = socket (AF_INET, SOCK_STREAM, 0);
+	s = socket (af, SOCK_STREAM, 0);
 	if (s < 0)
 	    err (1, "socket");
-	if (connect (s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
+	if (connect (s, addr, socket_sockaddr_size(addr)) < 0) {
 	    warn ("connect(%s)", hostname);
 	    close (s);
 	    continue;
 	}
-	errsock = socket (AF_INET, SOCK_STREAM, 0);
-	if (errsock < 0)
-	    err (1, "socket");
-	memset (&erraddr, 0, sizeof(erraddr));
-	erraddr.sin_family = AF_INET;
-	erraddr.sin_addr.s_addr = INADDR_ANY;
-	if (bind (errsock, (struct sockaddr *)&erraddr, sizeof(erraddr)) < 0)
-	    err (1, "bind");
+	if (do_errsock) {
+	    errsock = socket (af, SOCK_STREAM, 0);
+	    if (errsock < 0)
+		err (1, "socket");
+	    socket_set_any (erraddr, af);
+
+	    if (bind (errsock, erraddr, socket_sockaddr_size(erraddr)) < 0)
+		err (1, "bind");
+	} else
+	    errsock = -1;
     
-	return proto (s, errsock,
-		      hostname,
-		      local_user, remote_user,
-		      cmd, cmd_len);
+	ret = proto (s, errsock,
+		     hostname,
+		     local_user, remote_user,
+		     cmd, cmd_len, auth_func);
+	close (s);
+	return ret;
     }
     return 1;
 }
@@ -557,13 +717,14 @@ doit (char *hostname,
 #ifdef KRB4
 static int use_v4 = 0;
 #endif
-static int use_v5 = 1;
+static int use_v5 = 0;
 static int use_only_broken = 0;
 static int use_broken = 1;
 static char *port_str;
-static char *user;
+static const char *user;
 static int do_version;
 static int do_help;
+static int do_errsock = 1;
 
 struct getargs args[] = {
 #ifdef KRB4
@@ -578,12 +739,23 @@ struct getargs args[] = {
       NULL },
     { "encrypt", 'x', arg_flag,		&do_encrypt,	"Encrypt connection",
       NULL },
+    { "encrypt", 'z', arg_negative_flag,      &do_encrypt,
+      "Don't encrypt connection", NULL },
     { "forward", 'f', arg_flag,		&do_forward,	"Forward credentials",
       NULL },
+    { "forward", 'G', arg_negative_flag,&do_forward,	"Forward credentials",
+      NULL },
+    { "forwardable", 'F', arg_flag,	&do_forwardable,
+      "Forward forwardable credentials", NULL },
+    { "unique", 'u', arg_flag,	&do_unique_tkfile,
+      "Use unique remote tkfile", NULL },
+    { "tkfile", 'U', arg_string,  &unique_tkfile,
+      "Use that remote tkfile", NULL },
     { "port",	'p', arg_string,	&port_str,	"Use this port",
       "number-or-service" },
     { "user",	'l', arg_string,	&user,		"Run as this user",
       NULL },
+    { "stderr", 'e', arg_negative_flag, &do_errsock,	"don't open stderr"},
     { "version", 0,  arg_flag,		&do_version,	"Print version",
       NULL },
     { "help",	 0,  arg_flag,		&do_help,	NULL,
@@ -595,12 +767,13 @@ usage (int ret)
 {
     arg_printusage (args,
 		    sizeof(args) / sizeof(args[0]),
-		    "host command");
+		    NULL,
+		    "host [command]");
     exit (ret);
 }
 
 /*
- * main
+ *
  */
 
 int
@@ -613,8 +786,10 @@ main(int argc, char **argv)
     int ret = 1;
     char *cmd;
     size_t cmd_len;
-    struct passwd *pwd;
-    char *local_user;
+    const char *local_user;
+    char *host = NULL;
+    int host_index = -1;
+    int status; 
 
     priv_port1 = priv_port2 = IPPORT_RESERVED-1;
     priv_socket1 = rresvport(&priv_port1);
@@ -623,10 +798,43 @@ main(int argc, char **argv)
     
     set_progname (argv[0]);
 
+    if (argc >= 2 && argv[1][0] != '-') {
+	host = argv[host_index = 1];
+	optind = 1;
+    }
+    
+    status = krb5_init_context (&context);
+    if (status)
+        errx(1, "krb5_init_context failed: %u", status);
+      
+    do_forwardable = krb5_config_get_bool (context, NULL,
+					   "libdefaults",
+					   "forwardable",
+					   NULL);
+	
+    do_forward = krb5_config_get_bool (context, NULL,
+				       "libdefaults",
+				       "forward",
+				       NULL);
+
+    do_encrypt = krb5_config_get_bool (context, NULL,
+				       "libdefaults",
+				       "encrypt",
+				       NULL);
+
     if (getarg (args, sizeof(args) / sizeof(args[0]), argc, argv,
 		&optind))
 	usage (1);
 
+    if (do_forwardable)
+	do_forward = 1;
+
+    /* default to v5 */
+#ifdef KRB4
+    if(use_v4 == 0 && use_v5 == 0)
+#endif
+	use_v5 = 1;
+
     if (use_only_broken) {
 #ifdef KRB4
 	use_v4 = 0;
@@ -638,12 +846,36 @@ main(int argc, char **argv)
 	usage (0);
 
     if (do_version) {
-	printf ("%s (%s-%s)\n", __progname, PACKAGE, VERSION);
+	print_version (NULL);
 	return 0;
     }
 	
-    if (argc - optind < 2)
-	usage (1);
+    if (do_unique_tkfile && unique_tkfile != NULL)
+	errx (1, "Only one of -u and -U allowed.");
+
+    if (do_unique_tkfile)
+	strcpy(tkfile,"-u ");
+    else if (unique_tkfile != NULL) {
+	if (strchr(unique_tkfile,' ') != NULL) {
+	    warnx("Space is not allowed in tkfilename");
+	    usage(1);
+	}
+	do_unique_tkfile = 1;
+	snprintf (tkfile, sizeof(tkfile), "-U %s ", unique_tkfile);
+    }
+
+    if (host == NULL) {
+	if (argc - optind < 1)
+	    usage (1);
+	else
+	    host = argv[host_index = optind++];
+    }
+
+    if (optind == argc) {
+	argv[0] = "rlogin";
+	execvp ("rlogin", argv);
+	err (1, "execvp rlogin");
+    }
 
     if (port_str) {
 	struct servent *s = roken_getservbyname (port_str, "tcp");
@@ -660,15 +892,14 @@ main(int argc, char **argv)
 	}
     }
 
-    pwd = getpwuid (getuid());
-    if (pwd == NULL)
+    local_user = get_default_username ();
+    if (local_user == NULL)
 	errx (1, "who are you?");
-    local_user = pwd->pw_name;
 
     if (user == NULL)
 	user = local_user;
 
-    cmd_len = construct_command(&cmd, argc - optind - 1, argv + optind + 1);
+    cmd_len = construct_command(&cmd, argc - optind, argv + optind);
 
     /*
      * Try all different authentication methods
@@ -683,7 +914,9 @@ main(int argc, char **argv)
 	    tmp_port = krb5_getportbyname (context, "kshell", "tcp", 544);
 
 	auth_method = AUTH_KRB5;
-	ret = doit (argv[optind], user, local_user, tmp_port, cmd, cmd_len);
+	ret = doit (host, user, local_user, tmp_port, cmd, cmd_len,
+		    do_errsock,
+		    send_krb5_auth);
     }
 #ifdef KRB4
     if (ret && use_v4) {
@@ -697,7 +930,9 @@ main(int argc, char **argv)
 	    tmp_port = krb5_getportbyname (context, "kshell", "tcp", 544);
 
 	auth_method = AUTH_KRB4;
-	ret = doit (argv[optind], user, local_user, tmp_port, cmd, cmd_len);
+	ret = doit (host, user, local_user, tmp_port, cmd, cmd_len,
+		    do_errsock,
+		    send_krb4_auth);
     }
 #endif
     if (ret && use_broken) {
@@ -708,11 +943,11 @@ main(int argc, char **argv)
 	else
 	    tmp_port = krb5_getportbyname(context, "shell", "tcp", 514);
 	auth_method = AUTH_BROKEN;
-	ret = doit_broken (argc, argv, optind,
+	ret = doit_broken (argc, argv, host_index, host,
 			   user, local_user,
 			   tmp_port,
 			   priv_socket1,
-			   priv_socket2,
+			   do_errsock ? priv_socket2 : -1,
 			   cmd, cmd_len);
     }
     return ret;
