@@ -13,39 +13,9 @@
 #endif
 #endif
 
-int
-_krb5_extract_ticket __P((
-	krb5_context context,
-	krb5_kdc_rep *rep,
-	krb5_creds *creds,
-	krb5_keyblock *key,
-	krb5_const_pointer keyseed,
-	krb5_addresses *addrs,
-	unsigned nonce,
-	krb5_boolean allow_server_mismatch,
-	krb5_decrypt_proc decrypt_proc,
-	krb5_const_pointer decryptarg));
-
-void
-crc_init_table __P((void));
-
-u_int32_t
-crc_update __P((
-	char *p,
-	size_t len,
-	u_int32_t res));
-
-size_t
-k_get_int __P((
-	void *buffer,
-	unsigned long *value,
-	size_t size));
-
-size_t
-k_put_int __P((
-	void *buffer,
-	unsigned long value,
-	size_t size));
+#if !defined(__GNUC__) && !defined(__attribute__)
+#define __attribute__(x)
+#endif
 
 krb5_error_code
 krb524_convert_creds_kdc __P((
@@ -79,6 +49,26 @@ krb5_524_conv_principal __P((
 	char *instance,
 	char *realm));
 
+krb5_error_code
+krb5_abort __P((
+	krb5_context context,
+	krb5_error_code code,
+	const char *fmt,
+	...))
+    __attribute__ ((noreturn, format (printf, 3, 4)));
+
+krb5_error_code
+krb5_abortx __P((
+	krb5_context context,
+	const char *fmt,
+	...))
+    __attribute__ ((noreturn, format (printf, 2, 3)));
+
+krb5_error_code
+krb5_add_extra_addresses __P((
+	krb5_context context,
+	krb5_addresses *addresses));
+
 krb5_error_code
 krb5_addlog_dest __P((
 	krb5_context context,
@@ -97,7 +87,7 @@ krb5_addlog_func __P((
 
 krb5_error_code
 krb5_addr2sockaddr __P((
-	krb5_address *addr,
+	const krb5_address *addr,
 	struct sockaddr *sa,
 	int *sa_size,
 	int port));
@@ -120,6 +110,13 @@ krb5_address_search __P((
 	const krb5_address *addr,
 	const krb5_addresses *addrlist));
 
+krb5_error_code
+krb5_aname_to_localname __P((
+	krb5_context context,
+	krb5_const_principal aname,
+	size_t lnsize,
+	char *lname));
+
 krb5_error_code
 krb5_anyaddr __P((
 	int af,
@@ -127,6 +124,12 @@ krb5_anyaddr __P((
 	int *sa_size,
 	int port));
 
+krb5_error_code
+krb5_append_addresses __P((
+	krb5_context context,
+	krb5_addresses *dest,
+	const krb5_addresses *source));
+
 krb5_error_code
 krb5_auth_con_free __P((
 	krb5_context context,
@@ -187,6 +190,24 @@ krb5_auth_con_setflags __P((
 	krb5_auth_context auth_context,
 	int32_t flags));
 
+krb5_error_code
+krb5_auth_con_setkey __P((
+	krb5_context context,
+	krb5_auth_context auth_context,
+	krb5_keyblock *keyblock));
+
+krb5_error_code
+krb5_auth_con_setlocalsubkey __P((
+	krb5_context context,
+	krb5_auth_context auth_context,
+	krb5_keyblock *keyblock));
+
+krb5_error_code
+krb5_auth_con_setremotesubkey __P((
+	krb5_context context,
+	krb5_auth_context auth_context,
+	krb5_keyblock *keyblock));
+
 krb5_error_code
 krb5_auth_con_setuserkey __P((
 	krb5_context context,
@@ -206,10 +227,10 @@ krb5_auth_getcksumtype __P((
 	krb5_cksumtype *cksumtype));
 
 krb5_error_code
-krb5_auth_getenctype __P((
+krb5_auth_getkeytype __P((
 	krb5_context context,
 	krb5_auth_context auth_context,
-	krb5_enctype *etype));
+	krb5_keytype *keytype));
 
 krb5_error_code
 krb5_auth_getlocalseqnumber __P((
@@ -230,10 +251,10 @@ krb5_auth_setcksumtype __P((
 	krb5_cksumtype cksumtype));
 
 krb5_error_code
-krb5_auth_setenctype __P((
+krb5_auth_setkeytype __P((
 	krb5_context context,
 	krb5_auth_context auth_context,
-	krb5_enctype etype));
+	krb5_keytype keytype));
 
 krb5_error_code
 krb5_auth_setlocalseqnumber __P((
@@ -314,7 +335,7 @@ krb5_cc_default __P((
 	krb5_context context,
 	krb5_ccache *id));
 
-char*
+const char*
 krb5_cc_default_name __P((krb5_context context));
 
 krb5_error_code
@@ -334,7 +355,7 @@ krb5_cc_gen_new __P((
 	const krb5_cc_ops *ops,
 	krb5_ccache *id));
 
-char*
+const char*
 krb5_cc_get_name __P((
 	krb5_context context,
 	krb5_ccache id));
@@ -345,6 +366,16 @@ krb5_cc_get_principal __P((
 	krb5_ccache id,
 	krb5_principal *principal));
 
+const char*
+krb5_cc_get_type __P((
+	krb5_context context,
+	krb5_ccache id));
+
+krb5_error_code
+krb5_cc_get_version __P((
+	krb5_context context,
+	krb5_ccache id));
+
 krb5_error_code
 krb5_cc_initialize __P((
 	krb5_context context,
@@ -413,19 +444,17 @@ krb5_change_password __P((
 	krb5_data *result_string));
 
 krb5_boolean
-krb5_checksum_is_collision_proof __P((krb5_cksumtype ctype));
+krb5_checksum_is_collision_proof __P((
+	krb5_context context,
+	krb5_cksumtype type));
 
 krb5_boolean
-krb5_checksum_is_keyed __P((krb5_cksumtype ctype));
-
-krb5_error_code
-krb5_checksum_to_string __P((
+krb5_checksum_is_keyed __P((
 	krb5_context context,
-	krb5_cksumtype ctype,
-	char **string));
+	krb5_cksumtype type));
 
 krb5_error_code
-krb5_cksumsize __P((
+krb5_checksumsize __P((
 	krb5_context context,
 	krb5_cksumtype type,
 	size_t *size));
@@ -443,34 +472,55 @@ krb5_compare_creds __P((
 	krb5_creds *creds));
 
 krb5_error_code
-krb5_config_file_free __P((krb5_config_section *s));
+krb5_config_file_free __P((
+	krb5_context context,
+	krb5_config_section *s));
 
 void
 krb5_config_free_strings __P((char **strings));
 
 const void *
 krb5_config_get __P((
+	krb5_context context,
 	krb5_config_section *c,
 	int type,
 	...));
 
 krb5_boolean
 krb5_config_get_bool __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
+krb5_boolean
+krb5_config_get_bool_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	krb5_boolean def_value,
+	...));
+
 int
 krb5_config_get_int __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
+int
+krb5_config_get_int_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	int def_value,
+	...));
+
 const krb5_config_binding *
 krb5_config_get_list __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
 const void *
 krb5_config_get_next __P((
+	krb5_context context,
 	krb5_config_section *c,
 	krb5_config_binding **pointer,
 	int type,
@@ -478,19 +528,29 @@ krb5_config_get_next __P((
 
 const char *
 krb5_config_get_string __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
 char**
 krb5_config_get_strings __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
 int
 krb5_config_get_time __P((
+	krb5_context context,
 	krb5_config_section *c,
 	...));
 
+int
+krb5_config_get_time_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	int def_value,
+	...));
+
 krb5_error_code
 krb5_config_parse_file __P((
 	const char *fname,
@@ -498,27 +558,46 @@ krb5_config_parse_file __P((
 
 const void *
 krb5_config_vget __P((
+	krb5_context context,
 	krb5_config_section *c,
 	int type,
 	va_list args));
 
 krb5_boolean
 krb5_config_vget_bool __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
+krb5_boolean
+krb5_config_vget_bool_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	krb5_boolean def_value,
+	va_list args));
+
 int
 krb5_config_vget_int __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
+int
+krb5_config_vget_int_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	int def_value,
+	va_list args));
+
 const krb5_config_binding *
 krb5_config_vget_list __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
 const void *
 krb5_config_vget_next __P((
+	krb5_context context,
 	krb5_config_section *c,
 	krb5_config_binding **pointer,
 	int type,
@@ -526,19 +605,29 @@ krb5_config_vget_next __P((
 
 const char *
 krb5_config_vget_string __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
 char **
 krb5_config_vget_strings __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
 int
 krb5_config_vget_time __P((
+	krb5_context context,
 	krb5_config_section *c,
 	va_list args));
 
+int
+krb5_config_vget_time_default __P((
+	krb5_context context,
+	krb5_config_section *c,
+	int def_value,
+	va_list args));
+
 krb5_error_code
 krb5_copy_address __P((
 	krb5_context context,
@@ -596,12 +685,24 @@ krb5_copy_ticket __P((
 krb5_error_code
 krb5_create_checksum __P((
 	krb5_context context,
-	krb5_cksumtype type,
-	void *ptr,
+	krb5_crypto crypto,
+	unsigned usage_or_type,
+	void *data,
 	size_t len,
-	const krb5_keyblock *keyblock,
 	Checksum *result));
 
+krb5_error_code
+krb5_crypto_destroy __P((
+	krb5_context context,
+	krb5_crypto crypto));
+
+krb5_error_code
+krb5_crypto_init __P((
+	krb5_context context,
+	krb5_keyblock *key,
+	krb5_enctype etype,
+	krb5_crypto *crypto));
+
 krb5_error_code
 krb5_data_alloc __P((
 	krb5_data *p,
@@ -624,6 +725,62 @@ krb5_data_realloc __P((
 void
 krb5_data_zero __P((krb5_data *p));
 
+krb5_error_code
+krb5_decode_Authenticator __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	Authenticator *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_ETYPE_INFO __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	ETYPE_INFO *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_EncAPRepPart __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	EncAPRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_EncASRepPart __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	EncASRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_EncKrbCredPart __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	EncKrbCredPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_EncTGSRepPart __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	EncTGSRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_decode_EncTicketPart __P((
+	krb5_context context,
+	const void *data,
+	size_t length,
+	EncTicketPart *t,
+	size_t *len));
+
 krb5_error_code
 krb5_decode_ap_req __P((
 	krb5_context context,
@@ -633,17 +790,18 @@ krb5_decode_ap_req __P((
 krb5_error_code
 krb5_decrypt __P((
 	krb5_context context,
-	void *ptr,
+	krb5_crypto crypto,
+	unsigned usage,
+	void *data,
 	size_t len,
-	int etype,
-	const krb5_keyblock *keyblock,
 	krb5_data *result));
 
 krb5_error_code
 krb5_decrypt_EncryptedData __P((
 	krb5_context context,
+	krb5_crypto crypto,
+	unsigned usage,
 	EncryptedData *e,
-	const krb5_keyblock *keyblock,
 	krb5_data *result));
 
 krb5_error_code
@@ -651,7 +809,8 @@ krb5_decrypt_ticket __P((
 	krb5_context context,
 	Ticket *ticket,
 	krb5_keyblock *key,
-	EncTicketPart *out));
+	EncTicketPart *out,
+	krb5_flags flags));
 
 krb5_error_code
 krb5_domain_x500_decode __P((
@@ -667,56 +826,126 @@ krb5_domain_x500_encode __P((
 	int num_realms,
 	krb5_data *encoding));
 
+krb5_error_code
+krb5_encode_Authenticator __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	Authenticator *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_ETYPE_INFO __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	ETYPE_INFO *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_EncAPRepPart __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	EncAPRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_EncASRepPart __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	EncASRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_EncKrbCredPart __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	EncKrbCredPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_EncTGSRepPart __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	EncTGSRepPart *t,
+	size_t *len));
+
+krb5_error_code
+krb5_encode_EncTicketPart __P((
+	krb5_context context,
+	void *data,
+	size_t length,
+	EncTicketPart *t,
+	size_t *len));
+
 krb5_error_code
 krb5_encrypt __P((
 	krb5_context context,
-	void *ptr,
+	krb5_crypto crypto,
+	unsigned usage,
+	void *data,
 	size_t len,
-	int etype,
-	const krb5_keyblock *keyblock,
 	krb5_data *result));
 
 krb5_error_code
 krb5_encrypt_EncryptedData __P((
 	krb5_context context,
-	void *ptr,
+	krb5_crypto crypto,
+	unsigned usage,
+	void *data,
 	size_t len,
-	int etype,
 	int kvno,
-	const krb5_keyblock *keyblock,
 	EncryptedData *result));
 
+krb5_error_code
+krb5_enctype_to_keytype __P((
+	krb5_context context,
+	krb5_enctype etype,
+	krb5_keytype *keytype));
+
+krb5_error_code
+krb5_enctype_to_string __P((
+	krb5_context context,
+	krb5_enctype etype,
+	char **string));
+
+krb5_error_code
+krb5_enctype_valid __P((
+	krb5_context context,
+	krb5_enctype etype));
+
+krb5_boolean
+krb5_enctypes_comptible_keys __P((
+	krb5_context context,
+	krb5_enctype etype1,
+	krb5_enctype etype2));
+
 krb5_error_code
 krb5_err __P((
 	krb5_context context,
 	int eval,
 	krb5_error_code code,
 	const char *fmt,
-	...));
+	...))
+    __attribute__ ((noreturn, format (printf, 4, 5)));
 
 krb5_error_code
 krb5_errx __P((
 	krb5_context context,
 	int eval,
 	const char *fmt,
-	...));
+	...))
+    __attribute__ ((noreturn, format (printf, 3, 4)));
 
 krb5_error_code
-krb5_etype_to_keytype __P((
+krb5_expand_hostname __P((
 	krb5_context context,
-	krb5_enctype etype,
-	krb5_keytype *keytype));
-
-krb5_error_code
-krb5_etype_to_string __P((
-	krb5_context context,
-	krb5_enctype etype,
-	char **string));
-
-krb5_boolean
-krb5_etype_valid __P((
-	krb5_context context,
-	krb5_enctype etype));
+	const char *orig_hostname,
+	char **new_hostname));
 
 PA_DATA *
 krb5_find_padata __P((
@@ -748,6 +977,11 @@ krb5_free_authenticator __P((
 void
 krb5_free_context __P((krb5_context context));
 
+krb5_error_code
+krb5_free_cred_contents __P((
+	krb5_context context,
+	krb5_creds *c));
+
 krb5_error_code
 krb5_free_creds __P((
 	krb5_context context,
@@ -763,10 +997,20 @@ krb5_free_data __P((
 	krb5_context context,
 	krb5_data *p));
 
+void
+krb5_free_error __P((
+	krb5_context context,
+	krb5_error *error));
+
+void
+krb5_free_error_contents __P((
+	krb5_context context,
+	krb5_error *error));
+
 krb5_error_code
 krb5_free_host_realm __P((
 	krb5_context context,
-	char ** realmlist));
+	krb5_realm *realmlist));
 
 krb5_error_code
 krb5_free_kdc_rep __P((
@@ -793,11 +1037,27 @@ krb5_free_principal __P((
 	krb5_context context,
 	krb5_principal p));
 
+krb5_error_code
+krb5_free_salt __P((
+	krb5_context context,
+	krb5_salt salt));
+
 krb5_error_code
 krb5_free_ticket __P((
 	krb5_context context,
 	krb5_ticket *ticket));
 
+krb5_error_code
+krb5_fwd_tgt_creds __P((
+	krb5_context context,
+	krb5_auth_context auth_context,
+	const char *hostname,
+	krb5_principal client,
+	krb5_principal server,
+	krb5_ccache ccache,
+	int forwardable,
+	krb5_data *out_data));
+
 void
 krb5_generate_random_block __P((
 	void *buf,
@@ -806,7 +1066,7 @@ krb5_generate_random_block __P((
 krb5_error_code
 krb5_generate_random_keyblock __P((
 	krb5_context context,
-	krb5_keytype ktype,
+	krb5_enctype type,
 	krb5_keyblock *key));
 
 krb5_error_code
@@ -822,7 +1082,14 @@ krb5_generate_subkey __P((
 	krb5_keyblock **subkey));
 
 krb5_error_code
-krb5_get_all_client_addrs __P((krb5_addresses *res));
+krb5_get_all_client_addrs __P((
+	krb5_context context,
+	krb5_addresses *res));
+
+krb5_error_code
+krb5_get_all_server_addrs __P((
+	krb5_context context,
+	krb5_addresses *res));
 
 krb5_error_code
 krb5_get_cred_from_kdc __P((
@@ -862,12 +1129,22 @@ krb5_get_default_principal __P((
 krb5_error_code
 krb5_get_default_realm __P((
 	krb5_context context,
-	char **realm));
+	krb5_realm *realm));
 
 const char *
 krb5_get_err_text __P((
 	krb5_context context,
-	long code));
+	krb5_error_code code));
+
+krb5_error_code
+krb5_get_extra_addresses __P((
+	krb5_context context,
+	krb5_addresses *addresses));
+
+krb5_error_code
+krb5_get_fcache_version __P((
+	krb5_context context,
+	int *version));
 
 krb5_error_code
 krb5_get_forwarded_creds __P((
@@ -885,6 +1162,12 @@ krb5_get_host_realm __P((
 	const char *host,
 	krb5_realm **realms));
 
+krb5_error_code
+krb5_get_host_realm_int __P((
+	krb5_context context,
+	const char *host,
+	krb5_realm **realms));
+
 krb5_error_code
 krb5_get_in_cred __P((
 	krb5_context context,
@@ -892,6 +1175,7 @@ krb5_get_in_cred __P((
 	const krb5_addresses *addrs,
 	const krb5_enctype *etypes,
 	const krb5_preauthtype *ptypes,
+	const krb5_preauthdata *preauth,
 	krb5_key_proc key_proc,
 	krb5_const_pointer keyseed,
 	krb5_decrypt_proc decrypt_proc,
@@ -957,7 +1241,7 @@ krb5_get_init_creds_keytab __P((
 	krb5_principal client,
 	krb5_keytab keytab,
 	krb5_deltat start_time,
-	char *in_tkt_service,
+	const char *in_tkt_service,
 	krb5_get_init_creds_opt *options));
 
 void
@@ -1010,11 +1294,11 @@ krb5_get_init_creds_password __P((
 	krb5_context context,
 	krb5_creds *creds,
 	krb5_principal client,
-	char *password,
+	const char *password,
 	krb5_prompter_fct prompter,
 	void *data,
 	krb5_deltat start_time,
-	char *in_tkt_service,
+	const char *in_tkt_service,
 	krb5_get_init_creds_opt *options));
 
 krb5_error_code
@@ -1027,6 +1311,12 @@ krb5_get_kdc_cred __P((
 	krb5_creds *in_creds,
 	krb5_creds **out_creds ));
 
+krb5_error_code
+krb5_get_krb_admin_hst __P((
+	krb5_context context,
+	const krb5_realm *realm,
+	char ***hostlist));
+
 krb5_error_code
 krb5_get_krbhst __P((
 	krb5_context context,
@@ -1034,9 +1324,19 @@ krb5_get_krbhst __P((
 	char ***hostlist));
 
 krb5_error_code
-krb5_get_salt __P((
-	krb5_principal princ,
-	krb5_data *salt));
+krb5_get_pw_salt __P((
+	krb5_context context,
+	krb5_const_principal principal,
+	krb5_salt *salt));
+
+krb5_boolean
+krb5_get_use_admin_kdc __P((krb5_context context));
+
+size_t
+krb5_get_wrapped_length __P((
+	krb5_context context,
+	krb5_crypto crypto,
+	size_t data_len));
 
 int
 krb5_getportbyname __P((
@@ -1069,7 +1369,7 @@ krb5_error_code
 krb5_init_etype __P((
 	krb5_context context,
 	unsigned *len,
-	unsigned **val,
+	int **val,
 	const krb5_enctype *etypes));
 
 krb5_error_code
@@ -1087,27 +1387,24 @@ krb5_keyblock_key_proc __P((
 	krb5_keyblock **key));
 
 krb5_error_code
-krb5_keytype_to_cksumtype __P((
+krb5_keytab_key_proc __P((
 	krb5_context context,
-	krb5_keytype ktype,
-	krb5_cksumtype *ctype));
+	krb5_enctype enctype,
+	krb5_salt salt,
+	krb5_const_pointer keyseed,
+	krb5_keyblock **key));
 
 krb5_error_code
-krb5_keytype_to_etype __P((
-	krb5_context context,
-	krb5_keytype ktype,
-	krb5_enctype *etype));
-
-krb5_error_code
-krb5_keytype_to_etypes __P((
+krb5_keytype_to_enctypes __P((
 	krb5_context context,
 	krb5_keytype keytype,
-	krb5_enctype **etypes));
+	unsigned *len,
+	int **val));
 
 krb5_error_code
 krb5_keytype_to_string __P((
 	krb5_context context,
-	krb5_keytype ktype,
+	krb5_keytype keytype,
 	char **string));
 
 krb5_error_code
@@ -1136,7 +1433,7 @@ krb5_error_code
 krb5_kt_default_name __P((
 	krb5_context context,
 	char *name,
-	int namesize));
+	size_t namesize));
 
 krb5_error_code
 krb5_kt_end_seq_get __P((
@@ -1155,7 +1452,7 @@ krb5_kt_get_entry __P((
 	krb5_keytab id,
 	krb5_const_principal principal,
 	krb5_kvno kvno,
-	krb5_keytype keytype,
+	krb5_enctype enctype,
 	krb5_keytab_entry *entry));
 
 krb5_error_code
@@ -1163,7 +1460,7 @@ krb5_kt_get_name __P((
 	krb5_context context,
 	krb5_keytab keytab,
 	char *name,
-	int namesize));
+	size_t namesize));
 
 krb5_error_code
 krb5_kt_next_entry __P((
@@ -1178,9 +1475,20 @@ krb5_kt_read_service_key __P((
 	krb5_pointer keyprocarg,
 	krb5_principal principal,
 	krb5_kvno vno,
-	krb5_keytype keytype,
+	krb5_enctype enctype,
 	krb5_keyblock **key));
 
+krb5_error_code
+krb5_kt_register __P((
+	krb5_context context,
+	krb5_kt_ops *ops));
+
+krb5_error_code
+krb5_kt_remove_entry __P((
+	krb5_context context,
+	krb5_keytab id,
+	krb5_keytab_entry *entry));
+
 krb5_error_code
 krb5_kt_resolve __P((
 	krb5_context context,
@@ -1205,7 +1513,8 @@ krb5_log __P((
 	krb5_log_facility *fac,
 	int level,
 	const char *fmt,
-	...));
+	...))
+    __attribute__((format (printf, 4, 5)));
 
 krb5_error_code
 krb5_log_msg __P((
@@ -1214,7 +1523,14 @@ krb5_log_msg __P((
 	int level,
 	char **reply,
 	const char *fmt,
-	...));
+	...))
+    __attribute__((format (printf, 5, 6)));
+
+krb5_error_code
+krb5_make_addrport __P((
+	krb5_address **res,
+	const krb5_address *addr,
+	int16_t port));
 
 krb5_error_code
 krb5_make_principal __P((
@@ -1271,6 +1587,16 @@ krb5_mk_req_extended __P((
 	krb5_creds *in_creds,
 	krb5_data *outbuf));
 
+krb5_error_code
+krb5_mk_req_internal __P((
+	krb5_context context,
+	krb5_auth_context *auth_context,
+	const krb5_flags ap_req_options,
+	krb5_data *in_data,
+	krb5_creds *in_creds,
+	krb5_data *outbuf,
+	krb5_key_usage usage));
+
 krb5_error_code
 krb5_mk_safe __P((
 	krb5_context context,
@@ -1299,6 +1625,12 @@ krb5_openlog __P((
 	const char *program,
 	krb5_log_facility **fac));
 
+krb5_error_code
+krb5_parse_address __P((
+	krb5_context context,
+	const char *string,
+	krb5_addresses *addresses));
+
 krb5_error_code
 krb5_parse_name __P((
 	krb5_context context,
@@ -1308,8 +1640,8 @@ krb5_parse_name __P((
 krb5_error_code
 krb5_password_key_proc __P((
 	krb5_context context,
-	krb5_keytype type,
-	krb5_data *salt,
+	krb5_enctype type,
+	krb5_salt salt,
 	krb5_const_pointer keyseed,
 	krb5_keyblock **key));
 
@@ -1341,6 +1673,13 @@ krb5_principal_compare_any_realm __P((
 	krb5_const_principal princ1,
 	krb5_const_principal princ2));
 
+krb5_error_code
+krb5_print_address __P((
+	const krb5_address *addr,
+	char *str,
+	size_t len,
+	size_t *ret_len));
+
 int
 krb5_program_setup __P((
 	krb5_context *context,
@@ -1450,7 +1789,12 @@ krb5_ret_addrs __P((
 krb5_error_code
 krb5_ret_authdata __P((
 	krb5_storage *sp,
-	krb5_data *auth));
+	krb5_authdata *auth));
+
+krb5_error_code
+krb5_ret_creds __P((
+	krb5_storage *sp,
+	krb5_creds *creds));
 
 krb5_error_code
 krb5_ret_data __P((
@@ -1497,19 +1841,26 @@ krb5_ret_times __P((
 	krb5_storage *sp,
 	krb5_times *times));
 
+krb5_error_code
+krb5_salttype_to_string __P((
+	krb5_context context,
+	krb5_enctype etype,
+	krb5_salttype stype,
+	char **string));
+
 krb5_error_code
 krb5_sendauth __P((
 	krb5_context context,
 	krb5_auth_context *auth_context,
 	krb5_pointer p_fd,
-	char *appl_version,
+	const char *appl_version,
 	krb5_principal client,
 	krb5_principal server,
 	krb5_flags ap_req_options,
 	krb5_data *in_data,
 	krb5_creds *in_creds,
 	krb5_ccache ccache,
-	void **error,
+	krb5_error **ret_error,
 	krb5_ap_rep_enc_part **rep_result,
 	krb5_creds **out_creds));
 
@@ -1530,6 +1881,21 @@ krb5_set_default_realm __P((
 	krb5_context context,
 	char *realm));
 
+krb5_error_code
+krb5_set_extra_addresses __P((
+	krb5_context context,
+	krb5_addresses *addresses));
+
+krb5_error_code
+krb5_set_fcache_version __P((
+	krb5_context context,
+	int version));
+
+void
+krb5_set_use_admin_kdc __P((
+	krb5_context context,
+	krb5_boolean flag));
+
 krb5_error_code
 krb5_set_warn_dest __P((
 	krb5_context context,
@@ -1556,6 +1922,11 @@ krb5_sockaddr2address __P((
 	const struct sockaddr *sa,
 	krb5_address *addr));
 
+krb5_error_code
+krb5_sockaddr2port __P((
+	const struct sockaddr *sa,
+	int16_t *port));
+
 krb5_boolean
 krb5_sockaddr_uninteresting __P((const struct sockaddr *sa));
 
@@ -1565,12 +1936,20 @@ krb5_std_usage __P((
 	struct getargs *args,
 	int num_args));
 
+void
+krb5_storage_clear_flags __P((
+	krb5_storage *sp,
+	krb5_flags flags));
+
 krb5_storage *
 krb5_storage_emem __P((void));
 
 krb5_error_code
 krb5_storage_free __P((krb5_storage *sp));
 
+krb5_storage *
+krb5_storage_from_data __P((krb5_data *data));
+
 krb5_storage *
 krb5_storage_from_fd __P((int fd));
 
@@ -1579,6 +1958,16 @@ krb5_storage_from_mem __P((
 	void *buf,
 	size_t len));
 
+krb5_boolean
+krb5_storage_is_flags __P((
+	krb5_storage *sp,
+	krb5_flags flags));
+
+void
+krb5_storage_set_flags __P((
+	krb5_storage *sp,
+	krb5_flags flags));
+
 krb5_error_code
 krb5_storage_to_data __P((
 	krb5_storage *sp,
@@ -1597,7 +1986,12 @@ krb5_store_addrs __P((
 krb5_error_code
 krb5_store_authdata __P((
 	krb5_storage *sp,
-	krb5_data p));
+	krb5_authdata auth));
+
+krb5_error_code
+krb5_store_creds __P((
+	krb5_storage *sp,
+	krb5_creds *creds));
 
 krb5_error_code
 krb5_store_data __P((
@@ -1645,30 +2039,56 @@ krb5_store_times __P((
 	krb5_times times));
 
 krb5_error_code
-krb5_string_to_checksum __P((
+krb5_string_to_enctype __P((
 	krb5_context context,
 	const char *string,
-	krb5_cksumtype *ctype));
+	krb5_enctype *etype));
 
 krb5_error_code
 krb5_string_to_key __P((
-	const char *str,
-	krb5_data *salt,
-	krb5_keytype ktype,
+	krb5_context context,
+	krb5_enctype enctype,
+	const char *password,
+	krb5_principal principal,
 	krb5_keyblock *key));
 
 krb5_error_code
 krb5_string_to_key_data __P((
-	krb5_data *str,
-	krb5_data *salt,
-	krb5_keytype ktype,
+	krb5_context context,
+	krb5_enctype enctype,
+	krb5_data password,
+	krb5_principal principal,
+	krb5_keyblock *key));
+
+krb5_error_code
+krb5_string_to_key_data_salt __P((
+	krb5_context context,
+	krb5_enctype enctype,
+	krb5_data password,
+	krb5_salt salt,
+	krb5_keyblock *key));
+
+krb5_error_code
+krb5_string_to_key_derived __P((
+	krb5_context context,
+	const void *str,
+	size_t len,
+	krb5_enctype etype,
+	krb5_keyblock *key));
+
+krb5_error_code
+krb5_string_to_key_salt __P((
+	krb5_context context,
+	krb5_enctype enctype,
+	const char *password,
+	krb5_salt salt,
 	krb5_keyblock *key));
 
 krb5_error_code
 krb5_string_to_keytype __P((
 	krb5_context context,
 	const char *string,
-	krb5_keytype *ktype));
+	krb5_keytype *keytype));
 
 krb5_error_code
 krb5_timeofday __P((
@@ -1688,12 +2108,40 @@ krb5_unparse_name_fixed __P((
 	char *name,
 	size_t len));
 
+krb5_error_code
+krb5_unparse_name_fixed_short __P((
+	krb5_context context,
+	krb5_const_principal principal,
+	char *name,
+	size_t len));
+
+krb5_error_code
+krb5_unparse_name_short __P((
+	krb5_context context,
+	krb5_const_principal principal,
+	char **name));
+
 krb5_error_code
 krb5_us_timeofday __P((
 	krb5_context context,
 	int32_t *sec,
 	int32_t *usec));
 
+krb5_error_code
+krb5_vabort __P((
+	krb5_context context,
+	krb5_error_code code,
+	const char *fmt,
+	va_list ap))
+    __attribute__ ((noreturn, format (printf, 3, 0)));
+
+krb5_error_code
+krb5_vabortx __P((
+	krb5_context context,
+	const char *fmt,
+	va_list ap))
+    __attribute__ ((noreturn, format (printf, 2, 0)));
+
 krb5_error_code
 krb5_verify_ap_req __P((
 	krb5_context context,
@@ -1701,15 +2149,24 @@ krb5_verify_ap_req __P((
 	krb5_ap_req *ap_req,
 	krb5_const_principal server,
 	krb5_keyblock *keyblock,
+	krb5_flags flags,
 	krb5_flags *ap_req_options,
 	krb5_ticket **ticket));
 
+krb5_error_code
+krb5_verify_authenticator_checksum __P((
+	krb5_context context,
+	krb5_auth_context ac,
+	void *data,
+	size_t len));
+
 krb5_error_code
 krb5_verify_checksum __P((
 	krb5_context context,
-	void *ptr,
+	krb5_crypto crypto,
+	krb5_key_usage usage,
+	void *data,
 	size_t len,
-	const krb5_keyblock *keyblock,
 	Checksum *cksum));
 
 krb5_error_code
@@ -1735,7 +2192,7 @@ krb5_verify_user __P((
 	krb5_principal principal,
 	krb5_ccache ccache,
 	const char *password,
-	int secure,
+	krb5_boolean secure,
 	const char *service));
 
 krb5_error_code
@@ -1744,14 +2201,16 @@ krb5_verr __P((
 	int eval,
 	krb5_error_code code,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__ ((noreturn, format (printf, 4, 0)));
 
 krb5_error_code
 krb5_verrx __P((
 	krb5_context context,
 	int eval,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__ ((noreturn, format (printf, 3, 0)));
 
 krb5_error_code
 krb5_vlog __P((
@@ -1759,7 +2218,8 @@ krb5_vlog __P((
 	krb5_log_facility *fac,
 	int level,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__((format (printf, 4, 0)));
 
 krb5_error_code
 krb5_vlog_msg __P((
@@ -1768,33 +2228,38 @@ krb5_vlog_msg __P((
 	char **reply,
 	int level,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__((format (printf, 5, 0)));
 
 krb5_error_code
 krb5_vwarn __P((
 	krb5_context context,
 	krb5_error_code code,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__ ((format (printf, 3, 0)));
 
 krb5_error_code
 krb5_vwarnx __P((
 	krb5_context context,
 	const char *fmt,
-	va_list ap));
+	va_list ap))
+    __attribute__ ((format (printf, 2, 0)));
 
 krb5_error_code
 krb5_warn __P((
 	krb5_context context,
 	krb5_error_code code,
 	const char *fmt,
-	...));
+	...))
+    __attribute__ ((format (printf, 3, 4)));
 
 krb5_error_code
 krb5_warnx __P((
 	krb5_context context,
 	const char *fmt,
-	...));
+	...))
+    __attribute__ ((format (printf, 2, 3)));
 
 krb5_error_code
 krb5_write_message __P((
@@ -1802,6 +2267,9 @@ krb5_write_message __P((
 	krb5_pointer p_fd,
 	krb5_data *data));
 
+krb5_error_code
+krb5_xfree __P((void *ptr));
+
 krb5_error_code
 principalname2krb5_principal __P((
 	krb5_principal *principal,
